/*
 *  filter.c - Filtering mechanism
 *
 *  Copyright (C) 1998-2000 Hugo Haas & Etienne Bernard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <pwd.h>
#include <sys/types.h>
#include <stdlib.h>
#include <pthread.h>
#include <netdb.h>
#include <netinet/in.h>
#include <fnmatch.h>

#include "filter.h"
#include "resolv.h"
#include "configuration.h"

extern struct filterEntry *filter;

/* Parameters */
extern unsigned short resolveProtocols;
extern unsigned short icmpFormat;
extern unsigned short tcpFormat;
extern unsigned short udpFormat;
extern unsigned short logClosing;
extern unsigned short useIdent;

/* Mutexes and conditions */
pthread_mutex_t writeMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t readMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t writeCond = PTHREAD_COND_INITIALIZER;
int readers = 0;

/* Check if a number is in a particular range */
int checkRange(const uint16_t number, const long min, const long max) {
  /* No range. */
  if ((min == NOT_DEFINED) && (max == NOT_DEFINED))
    return 1;

  if (min == NOT_DEFINED) {
    if (number <= max)
      return 1;
  } else {
    if (max == NOT_DEFINED) {
      if (number >= min)
        return 1;
    } else {
      if ((number <= max) && (number >= min))
        return 1;
    }
  }

  return 0;
}

/* Check if address is described by address/mask */
int checkAddress(const uint32_t host, const uint32_t address, const uint32_t mask, char *name) {
  if (address == 0) {
    if (name == NULL)
      return 1;
    else
      return 0;
  }
  return ((address ^ host) & mask) == 0;
}

/* Check if a hostname matches a pattern */
int checkName(uint32_t address, const char *pattern) {
  char remoteHost[HOST_LENGTH];

  *remoteHost = '\0';
  hostLookup(remoteHost, address);

  if (!(pattern && *pattern)) return 1;

  if (!*remoteHost) return 0;

  if (!fnmatch(pattern, remoteHost, FNM_NOESCAPE)) return 1;

  return 0;
}


/* Add a reader to the filter */
void addReader(void) {
  pthread_mutex_lock(&readMutex);
  readers++;
  pthread_mutex_unlock(&readMutex);
}

/* Removes a reader from the filter */
void delReader(void) {
  pthread_mutex_lock(&readMutex);
  readers--;
  if (readers == 0)
    pthread_cond_signal(&writeCond);
  pthread_mutex_unlock(&readMutex);
}

/* Test if a packet has to be logged */
struct logInfo testLog(const uint32_t from, const uint32_t to, const uint16_t
                       type, const uint16_t srctype, const short protocol) {
  struct filterEntry *p;
  struct logInfo info;

  /* Add a reader to the filter */
  pthread_mutex_lock(&writeMutex);
  addReader();
  pthread_mutex_unlock(&writeMutex);

  for (p = filter; p != NULL; p = p->next) {
    /* Matching protocol? */
    if (p->protocol != protocol) continue;

    /* TCP or UDP */
    if (protocol != IPPROTO_ICMP) {
      if (checkRange(ntohs(type), p->logInfo.portRanges.dst.rangeMin,
                     p->logInfo.portRanges.dst.rangeMax) &&
          checkRange(ntohs(srctype), p->logInfo.portRanges.src.rangeMin,
                     p->logInfo.portRanges.src.rangeMax) &&
          checkAddress(from, p->fromDesc.address, p->fromDesc.mask,
                       p->fromDesc.hostName) &&
          (p->fromDesc.hostMask == NULL ||
           checkName(from, p->fromDesc.hostMask)) &&
          checkAddress(to, p->toDesc.address, p->toDesc.mask,
                        p->toDesc.hostName)) {
        delReader();
        info.log = p->log;
        info.ident = p->ident;
        info.resolve = p->resolve;
        info.logFormat = p->logFormat;
        info.logClosing = p->logClosing;
        return info;
      }
    } else {
      /* ICMP */
      if (((p->logInfo.icmpType == NO_ICMP_TYPE)
           || type == p->logInfo.icmpType) &&
          checkAddress(from, p->fromDesc.address, p->fromDesc.mask,
                       p->fromDesc.hostName) &&
          (p->fromDesc.hostMask == NULL ||
           checkName(from, p->fromDesc.hostMask)) &&
          checkAddress(to, p->toDesc.address, p->toDesc.mask,
                       p->toDesc.hostName)) {
        delReader();
        info.log = p->log;
        info.ident = p->ident;
        info.resolve = p->resolve;
        info.logFormat = p->logFormat;
        return info;
      }
    }
  }
  
  /* Default */
  delReader();

  info.log = 1;
  info.ident = useIdent;
  info.logClosing = logClosing;
  switch (protocol) {
  case IPPROTO_ICMP:
    info.resolve = resolveProtocols & GET_ICMP;
    info.logFormat = icmpFormat;
    break;
  case IPPROTO_TCP:
    info.resolve = resolveProtocols & GET_TCP;
    info.logFormat = tcpFormat;
    break;
  case IPPROTO_UDP:
    info.resolve = resolveProtocols & GET_UDP;
    info.logFormat = udpFormat;
    break;
  }
  return info;
}

/* Clear the filter */
void freeFilter(void) {
  struct filterEntry *p;

  /* Take mutexes on the filter */
  pthread_mutex_lock(&writeMutex);
  pthread_mutex_lock(&readMutex);
  while (readers > 0)
    pthread_cond_wait(&writeCond, &readMutex);

  while ((p = filter) != NULL) {
    filter = filter->next;
    if (p->fromDesc.hostName != NULL) free(p->fromDesc.hostName);
    if (p->fromDesc.hostMask != NULL) free(p->fromDesc.hostMask);
    if (p->toDesc.hostName != NULL) free(p->toDesc.hostName);
    if (p->toDesc.hostMask != NULL) free(p->toDesc.hostMask);
    free(p);
  }

  /* Release the mutexes */
  pthread_mutex_unlock(&writeMutex);
  pthread_mutex_unlock(&readMutex);
}

/* Do DNS queries for entries in the filter */
void refreshFilter() {
  struct filterEntry *p;
  struct hostent * host;

  /* Get mutexes on the filter */
  pthread_mutex_lock(&writeMutex);
  pthread_mutex_lock(&readMutex);
  while (readers > 0)
    pthread_cond_wait(&writeCond, &readMutex);

  for (p = filter; p != NULL; p = p->next) {
    if (p->fromDesc.hostName != NULL) {
      if ((host = gethostbyname((const char *)p->fromDesc.hostName)) == 0) {
        p->fromDesc.address = 0;
        p->fromDesc.mask = 0;
      }
      else {
        memcpy(&(p->fromDesc.address), host->h_addr, host->h_length);
        p->fromDesc.mask = 0xffffffff;
      }
    }
  }

  /* Release the mutexes */
  pthread_mutex_unlock(&readMutex);
  pthread_mutex_unlock(&writeMutex);
}
