/*
 *  pcap.c - interface to libpcap
 *
 *  Copyright (C) 1999-2000 Hugo Haas & Etienne Bernard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "defines.h"

#ifdef __OS_LINUX__
#	define __USE_BSD 1
#	define __FAVOR_BSD 1
#endif

#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <string.h>
#include <pcap.h>
#include <pwd.h>
#include <sys/types.h>

#include "pcap.h"
#include "log.h"
#include "configuration.h"
#include "filter.h"
#include "logging.h"

/* Log devices */
extern logDevice *mainLog;
extern logDevice *icmpLog;
extern logDevice *tcpLog;
extern logDevice *udpLog;

extern unsigned short logProtocols;

/* Compute checksum from an IP packet.
   Copyright (c) 1999 Robert Cheramy <tibob@via.ecp.fr>
                  and Andres Krapf <dae@via.ecp.fr> */
int ipChecksum(u_char *iphead, int size)
{
  unsigned long int sum = 0;
  int i;

  for (i = (size * 4) - 2; i > -1; i-=2) {
    sum += iphead[i] + 256 * iphead[i+1];
  }
  sum = (sum & 0xffff) + (sum >> 16);
  /* mmm I'm not sure the "(sum >> 16)" is usfull, but better be sure */
  sum += (sum >> 16) + 1;
  return (sum & 0xffff);
}

/* Get offset from datalink */
int getOffset(pcap_t *desc) {
  int datalink = pcap_datalink(desc);

  switch(datalink) {
  case DLT_EN10MB:
    return 14;
  case DLT_NULL:
  case DLT_PPP:
    return 4;
  case DLT_SLIP:
    return 16;
#ifndef __OS_OPENBSD__
  case DLT_RAW:
    return 0;
  case DLT_SLIP_BSDOS:
  case DLT_PPP_BSDOS:
    return 24;
#endif /* __OS_OPENBSD__ */
  case DLT_ATM_RFC1483:
    return 8;
  case DLT_IEEE802:
    return 22;
  default:
    mainLog->doLog(mainLog, "Unknown datalink type (%d)", datalink);
    return 0;
  }
}

/* Main interface function */
void interfaceThread(struct interfaceThreadParams *params) {
  char buffer[PCAP_ERRBUF_SIZE], *filter;
  struct bpf_program filt;
  struct passwd *user;
  char *interface;
  struct in_addr *ipAddresses;
  unsigned short countIP;
  pcap_t *pcapDesc;
  int i, pcapOffset;
  u_char *packetRaw;
  struct pcap_pkthdr dummy;
  struct ip *ip;
  struct icmp *icmp;
  struct udphdr *udp;
  struct tcphdr *tcp;
  uint32_t from , to;
  struct logInfo info;

  /* Get parameters */
  user = params->user;
  interface = params->name;
  countIP = params->countIP;
  ipAddresses = (struct in_addr *)malloc(countIP * sizeof(struct in_addr));
  for (i = 0; i < countIP; i++)
    ipAddresses[i] = params->ip[i];

#ifdef DEBUG
  mainLog->doLog(mainLog, "Starting thread on interface %s", interface);
#endif

  free(params->ip);
  free(params);

  /* Enable cancelation */
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  /* Open pcap descriptor */
  pcapDesc = pcap_open_live(interface, 68, 0, 0, buffer);
  if (!pcapDesc) {
    mainLog->doLog(mainLog, "Error opening interface. Stopping thread...");
    pthread_exit(NULL);
  }

#ifdef __OS_LINUX__
  /* Switch user */
  if (user) setuid(user->pw_uid);
#endif

  /* Create packet filter */
  *buffer = '\0';
  if (logProtocols & GET_ICMP) strcat(buffer, " or icmp");
  if (logProtocols & GET_TCP) strcat(buffer, " or tcp");
  if (logProtocols & GET_UDP) strcat(buffer, " or udp");
  strcat(buffer, ") and dst host (");
  strcat(buffer, inet_ntoa(ipAddresses[0]));
  for (i = 1; i < countIP; i++) {
    strcat(buffer, " or ");
    strcat(buffer, inet_ntoa(ipAddresses[i]));
  }
  strcat(buffer, ")");
  filter = buffer + 1;
  filter = strchr(filter, ' ');
  *filter = '(';

#ifdef DEBUG
  mainLog->doLog(mainLog, "Filter on interface %s: %s", interface, filter);
#endif

  /* Get offset */
  pcapOffset = getOffset(pcapDesc);

  /* Compile packet filter */
  if (pcap_compile(pcapDesc, &filt, filter, 1, 0) == -1) {
    mainLog->doLog(mainLog, "WARNING: Unable to compile filter on interface %s.", interface);
    pthread_exit(NULL);
  }

  /* Get packets */
  while (1) {
    info.log = 0;
    packetRaw = (u_char *)pcap_next(pcapDesc, &dummy);
    if (!packetRaw) continue;
    packetRaw += pcapOffset;
    ip = (struct ip *)packetRaw;

    /* Is it a valid packet? */
    if (ip->ip_v != 4 ||
        ipChecksum(packetRaw, ip->ip_hl) != 0) {
      /* Not a valid IP v4 packet */
      continue;
    }
    
    from = ip->ip_src.s_addr;
    to = ip->ip_dst.s_addr;

    /* Check the packet :
     * - if it comes from us, ignore it
     * - if it is for us, log it */
    for (i = 0; i < countIP; i++)
      if (from == ipAddresses[i].s_addr) goto drop;

    for (i = 0; i < countIP; i++)
      if (to == ipAddresses[i].s_addr) goto cont;

  drop:
    continue;

  cont:

    /* Now, we have to decide if we send the packet to the logging
       thread. */
    if (ip->ip_p == IPPROTO_ICMP && logProtocols & GET_ICMP) {
      icmp = (struct icmp *)(packetRaw + (ip->ip_hl << 2));
      info = testLog(from, to, icmp->icmp_type, 0, IPPROTO_ICMP);
      info.protocol = IPPROTO_ICMP;
      info.from = from;
      info.to = to;
      info.srcPort = icmp->icmp_type;
      info.dstPort = icmp->icmp_code;
      if (icmpLog)
        info.logDevice = icmpLog;
      else
        info.logDevice = mainLog;
    }

    if (ip->ip_p == IPPROTO_UDP && logProtocols & GET_UDP) {
      udp = (struct udphdr *)(packetRaw + (ip->ip_hl << 2));
      info = testLog(from, to, udp->uh_dport, udp->uh_sport, IPPROTO_UDP);
      info.protocol = IPPROTO_UDP;
      info.from = from;
      info.to = to;
      info.srcPort = udp->uh_sport;
      info.dstPort = udp->uh_dport;
      if (udpLog)
        info.logDevice = udpLog;
      else
        info.logDevice = mainLog;
    }

    if (ip->ip_p == IPPROTO_TCP && logProtocols & GET_TCP) {
      tcp = (struct tcphdr *)(packetRaw + (ip->ip_hl << 2));
      if ((tcp->th_flags & TH_SYN && !(tcp->th_flags & TH_ACK)) || tcp->th_flags & TH_FIN) {
        info = testLog(from, to, tcp->th_dport, tcp->th_sport, IPPROTO_TCP);
        info.protocol = IPPROTO_TCP;
        info.from = from;
        info.to = to;
        info.srcPort = tcp->th_sport;
        info.dstPort = tcp->th_dport;
        if (tcpLog)
          info.logDevice = tcpLog;
        else
          info.logDevice = mainLog;
        /* Should we log this closing connection ? */
        if (tcp->th_flags & TH_FIN) {
          info.closing = 1;
          info.log = info.logClosing;
        } else
          info.closing = 0;
      }
    }

    if (info.log) {
      /* Record length of the IP header */
      info.ihl = ip->ip_hl;
      recordPacket(info);
    }
  }
}
