/*
 *  utils.c - Utility functions
 *
 *  Copyright (C) 1999-2000 Hugo Haas & Etienne Bernard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>
#ifdef __OS_LINUX__
#	if (__GNU_LIBRARY__ < 6)
#		include "icmpfix.h"
#	endif
#endif
#include <netdb.h>
#include <stdio.h>

#include "defines.h"
#include "utils.h"
#include "resolv.h"

/* Missing ICMP definitions */
#ifndef ICMP_UNREACH_NET_UNKNOWN
#	define ICMP_UNREACH_NET_UNKNOWN		6
#endif
#ifndef ICMP_UNREACH_HOST_UNKNOWN
#	define ICMP_UNREACH_HOST_UNKNOWN	7
#endif
#ifndef ICMP_UNREACH_ISOLATED
#	define ICMP_UNREACH_ISOLATED		8
#endif
#ifndef ICMP_UNREACH_NET_PROHIB
#	define ICMP_UNREACH_NET_PROHIB		9
#endif
#ifndef ICMP_UNREACH_HOST_PROHIB
#	define ICMP_UNREACH_HOST_PROHIB		10
#endif
#ifndef ICMP_UNREACH_TOSNET
#	define ICMP_UNREACH_TOSNET		11
#endif
#ifndef ICMP_UNREACH_TOSHOST
#	define ICMP_UNREACH_TOSHOST		12
#endif
#ifndef ICMP_UNREACH_FILTER_PROHIB
#	define ICMP_UNREACH_FILTER_PROHIB	13
#endif
#ifndef ICMP_UNREACH_HOST_PRECEDENCE
#	define ICMP_UNREACH_HOST_PRECEDENCE	14
#endif
#ifndef ICMP_UNREACH_PRECEDENCE_CUTOFF
#	define ICMP_UNREACH_PRECEDENCE_CUTOFF	15
#endif
#ifndef ICMP_SOURCE_QUENCH
#	define ICMP_SOURCE_QUENCH		4
#endif
#ifndef ICMP_TIME_EXCEEDED
#	define ICMP_TIME_EXCEEDED		11
#endif
#ifndef ICMP_PARAMETERPROB
#	define ICMP_PARAMETERPROB		12
#endif
#ifndef ICMP_TIMESTAMP
#	define ICMP_TIMESTAMP			13
#endif
#ifndef ICMP_TIMESTAMPREPLY
#	define ICMP_TIMESTAMPREPLY		14
#endif
#ifndef ICMP_INFO_REQUEST
#	define ICMP_INFO_REQUEST		15
#endif
#ifndef ICMP_INFO_REPLY
#	define ICMP_INFO_REPLY			16
#endif
#ifndef ICMP_ADDRESS
#	define ICMP_ADDRESS			17
#endif
#ifndef ICMP_ADDRESSREPLY
#	define ICMP_ADDRESSREPLY		18
#endif

#define ICMP_ROUTER_ADVERT       9
#define ICMP_ROUTER_SOLICITATION 10

/* Print IP address, or host name */
void hostPrint(char *buffer, uint32_t ip, unsigned short resolve) {
  char remoteHostname[HOST_LENGTH];
  struct in_addr i;

  i.s_addr = ip;
  *remoteHostname = '\0';

  if (resolve && hostLookup(remoteHostname, ip))
    snprintf(buffer, HOST_LENGTH, "%s [%.20s]", remoteHostname, inet_ntoa(i));
  else
    snprintf(buffer, 21, "%.20s", inet_ntoa(i));
}

/* Get strings for ICMP type and code */
void icmpTypeLookup(char *string1, char *string2, uint16_t type, uint16_t code) {

  *string2 = '\0';

  switch(type) {
  case ICMP_ECHOREPLY: /* Echo reply */
    strcpy(string1,"echo reply");
    return;
  
  case ICMP_UNREACH: /* Destination unreachable */
    strcpy(string1,"destination unreachable");
    switch(code) {
    case ICMP_UNREACH_NET:
      strcpy(string2," - bad network");
      return;
    case ICMP_UNREACH_HOST:
      strcpy(string2," - bad host");
      return;
    case ICMP_UNREACH_PROTOCOL:
      strcpy(string2," - bad protocol");
      return;
    case ICMP_UNREACH_PORT:
      strcpy(string2," - bad port");
      return;
    case ICMP_UNREACH_NEEDFRAG:
      strcpy(string2," - IP_DF caused drop");
      return;
    case ICMP_UNREACH_SRCFAIL:
      strcpy(string2," - source route failed");
      return;
    case ICMP_UNREACH_NET_UNKNOWN:
      strcpy(string2," - unknown network");
      return;
    case ICMP_UNREACH_HOST_UNKNOWN:
      strcpy(string2," - unknown host");
      return;
    case ICMP_UNREACH_ISOLATED:
      strcpy(string2," - src host isolated");
      return;
    case ICMP_UNREACH_NET_PROHIB:
      strcpy(string2," - network denied");
      return;
    case ICMP_UNREACH_HOST_PROHIB:
      strcpy(string2," - host denied");
      return;
    case ICMP_UNREACH_TOSNET:
      strcpy(string2," - bad tos for network");
      return;
    case ICMP_UNREACH_TOSHOST:
      strcpy(string2," - bad tos for host");
      return;
    case ICMP_UNREACH_FILTER_PROHIB:
      strcpy(string2," - packed filtered");
      return;
    case ICMP_UNREACH_HOST_PRECEDENCE:
      strcpy(string2," - host precedence violation");
      return;
    case ICMP_UNREACH_PRECEDENCE_CUTOFF:
      strcpy(string2," - precedence cutoff");
      return;
    default:
      return;
    }

  case ICMP_SOURCE_QUENCH: /* Source quench */
    strcpy(string1,"source quench");
    return;
    
  case ICMP_REDIRECT: /* Redirection */
    strcpy(string1,"redirect");
    switch(code) {
    case ICMP_REDIRECT_NET:
      strcpy(string2," - for network");
      return;
    case ICMP_REDIRECT_HOST:
      strcpy(string2," - for host");
      return;
    case ICMP_REDIRECT_TOSNET:
      strcpy(string2," - for tos and net");
      return;
    case ICMP_REDIRECT_TOSHOST:
      strcpy(string2," - for tos and host");
      return;
    default:
      return;
    }

  case ICMP_ECHO: /* Echo */
    strcpy(string1,"echo request");
    return;

  case ICMP_ROUTER_ADVERT: /* Router advertisement */
    strcpy(string1,"router advertisement");
    return;

  case ICMP_ROUTER_SOLICITATION: /* Router solicitation */
    strcpy(string1,"router solicitation");
    return;

  case ICMP_TIME_EXCEEDED: /* Time exceeded */
    strcpy(string1,"time exceeded");
    return;

  case ICMP_PARAMETERPROB: /* Parameter problem */
    strcpy(string1,"parameter problem");
    return;

  case ICMP_TIMESTAMP: /* Timestamp request */
    strcpy(string1,"timestamp request");
    return;

  case ICMP_TIMESTAMPREPLY: /* Timestamp reply */
    strcpy(string1,"timestamp reply");
    return;
  
  case ICMP_INFO_REQUEST: /* Information request */
    strcpy(string1,"information request");
    return;

  case ICMP_INFO_REPLY: /* Information reply */
    strcpy(string1,"information reply");
    return;

  case ICMP_ADDRESS: /* Address mask request */
    strcpy(string1,"address mask request");
    return;

  case ICMP_ADDRESSREPLY: /* Address mask reply */
    strcpy(string1,"address mask reply");
    return;

  default: /* Default */
    strcpy(string1,"unknown");
  }
}

/* Print details about a packet.
 * Based on Steffen Ullrich's (ccrlphr@xensei.com) code
 * " (xxx.xxx.xxx.xxx:xxxxx->xxx.xxx.xxx.xxx:xxxxx)" if ports are non-zero
 * " (xxx.xxx.xxx.xxx->xxx.xxx.xxx.xxx)" if ports are zero
 */
void getDetails(char *details, uint32_t srcAddr, uint16_t srcPort, uint32_t dstAddr, uint16_t dstPort) {
  int sz;
  struct in_addr i;

  i.s_addr = srcAddr;
  if ((srcPort == 0) && (dstPort == 0)) {
    sz = snprintf(details, DETAILS_LENGTH, " (%s->", inet_ntoa(i));
    i.s_addr = dstAddr;
    snprintf(details+sz, DETAILS_LENGTH-sz, "%s)", inet_ntoa(i));
  }
  else {
    sz = snprintf(details, DETAILS_LENGTH, " (%s:%d->", inet_ntoa(i), ntohs(srcPort));
    i.s_addr = dstAddr;
    snprintf(details+sz, DETAILS_LENGTH-sz, "%s:%d)", inet_ntoa(i), ntohs(dstPort));
  }
}

/* Get a service name for a specified protocol */
void serviceLookup(char *proto, char *service, uint16_t port) {
  struct servent *se;

  se = getservbyport(port, proto);

  if (se == NULL)
    snprintf(service, SERVICE_LENGTH, "port %d", ntohs(port));
  else
    snprintf(service, SERVICE_LENGTH, "%s", se->s_name);
}
